from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .data import *
from .optimizers import *
from .structure import *
from .transforms import *

_allowed_symbols = ['laplace', 'svi',
                     'MFSVI', 'Laplace']


import numpy as np


class CG:

    def __init__(self, cg_prod=None, tol=1e-3):
        self.cg_prod = cg_prod
        self.tol = tol

    def cg(self, A, b, x = None):

        n = len(b)
        if not x:
            x = np.ones(n)

        r = self.cg_prod(A, x) - b
        p = - r
        r_k_norm = np.dot(r, r)
        for i in range(2 * n):
            Ap = self.cg_prod(A, p)
            alpha = r_k_norm / np.dot(p, Ap)
            x += alpha * p
            r += alpha * Ap
            r_kplus1_norm = np.dot(r, r)
            beta = r_kplus1_norm / r_k_norm
            r_k_norm = r_kplus1_norm
            if r_kplus1_norm < self.tol:
                break
            p = beta * p - r

        return x

class Adam:

    def __init__(self, step_size = 1e-3, b1 = .9, b2 = .999, eps = 0.1):
        self.step_size = step_size
        self.b1 = b1
        self.b2 = b2
        self.eps = eps

    def step(self, var_and_grad, m, v, t):
        """
        Adapted from autograd.misc.optimizers
        Args:
            S_grads ():
            mu_grads ():
            kern_grads ():
            step_size ():
            b1 ():
            b2 ():
            eps ():

        Returns:

        """

        var, grad = var_and_grad
        if m is None:
            m = np.zeros(len(var))
        if v is None:
            v = np.zeros(len(var))

        m = self.b1* m + (1-self.b1)*grad
        v = self.b2*v + (1-self.b2)*np.square(v)
        alpha_t = self.step_size*np.sqrt(1-self.b2**t)/(1-self.b1**t)
        var = var + alpha_t*m/(np.sqrt(v) + self.eps)

        return var, m, v

class RMSProp:

    def __init__(self, step_size=1e-3, gamma=0.9, eps=0.1):
        self.step_size = step_size
        self.gamma = gamma
        self.eps = eps

    def step(self, var_and_grad, avg_sq_grad = None):

        """Root mean squared prop: See Adagrad paper for details."""
        var, grad = var_and_grad
        if avg_sq_grad is None:
            avg_sq_grad = np.ones(len(var))
        avg_sq_grad = avg_sq_grad * self.gamma + grad ** 2 * (1 - self.gamma)
        var = var + self.step_size * grad / (np.sqrt(avg_sq_grad) + self.eps)

        return var, avg_sq_grad


import autograd.numpy as np
from gp3.utils.transforms import softplus, inv_softplus

class RBF:

    def __init__(self, lengthscale, variance, noise):

        self.params = self.pack_params(lengthscale, variance, noise)

    def eval(self, params, X, X2 = None):
        """
        RBF (squared exponential) kernel with softplus transform of lengthscale and variance
        Args:
            params (np.array): lengthscale and variance (in order)
            X (): first X
            X2 (): second X (if there is one, otherwise just eval on X)

        Returns:s

        """

        ls, var, noise = self.unpack_params(params)

        if X2 is None:
            X2 = X

        delta = np.expand_dims(X / ls, 1) -\
                np.expand_dims(X2 / ls, 0)

        return var * np.exp(-0.5 * np.sum(delta ** 2, axis=2)) +\
               np.diag(np.ones(X.shape[0]))*noise

    def unpack_params(self, params):

        return softplus(params[0]), softplus(params[1]), softplus(params[2])

    def pack_params(self, lengthscale, variance, noise):

        return inv_softplus(np.array([lengthscale, variance, noise]))

class DeepRBF:
    """
    Inspired by autograd's Bayesian neural net example
    """

    def __init__(self, layer_sizes, params):

        self.params = params
        self.shapes = list(zip(layer_sizes[:-1], layer_sizes[1:]))

    def eval(self, params, X, X2 = None):
        """
        RBF (squared exponential) kernel with softplus
         transform of lengthscale and variance
        Args:
            params (np.array): lengthscale and variance (in order)
            X (): first X
            X2 (): second X (if there is one, otherwise just eval on X)

        Returns:s

        """

        ls, var, weights = self.unpack_params(params)

        if X2 is None:
            X2 = X

        X_nn = self.nn_predict(weights, X)
        X2_nn = self.nn_predict(weights, X2)

        delta = np.expand_dims(X_nn / ls, 1) -\
                np.expand_dims(X2_nn / ls, 0)

        return var * np.exp(-0.5 * np.sum(delta ** 2, axis=2))

    def unpack_params(self, params):

        return softplus(params[0]), softplus(params[1]), params[2:]

    def unpack_layers(self, weights):

        num_weight_sets = len(weights)
        for m, n in self.shapes:
            yield weights[:, :m * n].reshape((num_weight_sets, m, n)), \
                  weights[:, m * n:m * n + n].reshape((num_weight_sets, 1, n))
            weights = weights[:, (m + 1) * n:]

    def nn_predict(self, weights, inputs):

        inputs = np.expand_dims(inputs, 0)
        outputs = None
        for W, b in self.unpack_layers(weights):
            outputs = np.einsum('mnd,mdo->mno', inputs, W) + b
            inputs = softplus(outputs)

        return outputs

class SpectralMixture:

    def __init__(self, params):
        self.params = params



from .laplace import *
from .svi import *

__allowed_symbols = ['sim_f', 'sim_X', 'poisson_draw',
                     'sim_X_equispaced', 'poisson_draw',
                     'rand_partial_grid',
                     'fill_grid', 'CG', 'Adam',
                     'kron', 'kron_mvp', 'kron_list'
                     'softplus', 'inv_softplus']


import numpy as np
from functools import reduce

def kron(A, B):
    """
    Kronecker product of two matrices
    Args:
        A (np.array): first matrix for kronecker product
        B (np.array): second matrix

    Returns: kronecker product of A and B

    """

    n_col = A.shape[1] * B.shape[1]
    out = np.zeros([0, n_col])

    for i in range(A.shape[0]):

        row = np.zeros([B.shape[0], 0])

        for j in range(A.shape[1]):
            row = np.concatenate([row, A[i, j]* B], 1)

        out = np.concatenate([out, row], 0)

    return out


def kron_list(matrices):
    """
    Kronecker product of a list of matrices
    Args:
        matrices (list of np.array): list of matrices

    Returns:

    """
    return reduce(kron, matrices)


def kron_mvp(Ks, v):
    """
    Matrix vector product using Kronecker structure
    Args:
        Ks (list of np.array): list of matrices
        of K
        v (np.array): vector to multiply K by

    Returns: matrix vector product of K and v

    """

    mvp = v

    for k in reversed(Ks):
        mvp = np.reshape(mvp, [k.shape[0], -1])
        mvp = np.dot(k, mvp).T

    return mvp.flatten()

def kron_list_diag(Ks):

    diag = np.hstack([ii * np.diag(Ks[len(Ks)-1])
                      for ii in np.diag(Ks[len(Ks)-2])])

    for i in reversed(range(len(Ks[:-2]))):
        diag = np.hstack([ii * diag for ii in np.diag(Ks[i])])

    return diag


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import inference
from . import likelihoods
from . import utils
from . import kernels

_allowed_symbols = ['inference', 'likelihoods', 'utils', 'kernels']



import autograd.numpy as np
from autograd import elementwise_grad as egrad, jacobian
from gp3.utils.structure import kron_mvp, kron_list_diag
from scipy.linalg import toeplitz
from gp3.utils.optimizers import Adam
from tqdm import trange, tqdm_notebook
from copy import deepcopy
import scipy

"""
Stochastic Variational Inference for Gaussian Processes with
 Non-Gaussian Likelihoods
"""


class SVIBase(object):

    def __init__(self, kernel, likelihood, X, y, mu=None, obs_idx=None,
                 opt_kernel=False, max_grad=10.,
                 optimizer=Adam()):

        self.X = X
        self.y = y
        self.n, self.d = self.X.shape
        self.X_dims = [np.expand_dims(np.unique(X[:, i]), 1)
                       for i in range(self.d)]
        if mu is None:
            self.mu = np.zeros(self.n)
        else:
            self.mu = mu
        self.obs_idx = obs_idx
        self.max_grad = max_grad
        self.init_Ks(kernel, opt_kernel)
        self.elbos = []
        self.q_mu = self.mu

        self.likelihood = likelihood
        self.likelihood_opt = egrad(self.likelihood.log_like)
        self.optimizer = optimizer

    def init_Ks(self, kernel, opt_kernel):

        self.kernel = kernel
        self.Ks, self.K_invs = self.construct_Ks()
        self.k_inv_diag = kron_list_diag(self.K_invs)
        self.det_K = self.log_det_K()
        self.opt_kernel = opt_kernel
        if opt_kernel:
            self.kernel_opt = jacobian(self.kernel.eval)

    def log_det_K(self, Ks=None):
        """
        Log determinant of prior covariance
        Returns: log determinant
        """
        if Ks is None:
            Ks = self.Ks

        log_det = 0.

        for K in Ks:
            rank_d = self.n / K.shape[0]
            det = np.linalg.slogdet(K)[1]
            log_det += rank_d * det

        return log_det

    def construct_Ks(self, kernel=None):
        """
        Constructs kronecker-decomposed kernel matrix
        Args:
            kernel (): kernel (if not using kernel passed in constructor)
        Returns: Rist of kernel evaluated at each dimension
        """

        if kernel is None:
            kernel = self.kernel

        Ks = [kernel.eval(kernel.params, X_dim)
              for X_dim in self.X_dims]
        K_invs = [np.linalg.inv(K) for K in Ks]

        return Ks, K_invs

    def predict(self):
        """
        GP predictions
        Returns: predictions
        """
        Ks = []
        for i in range(self.X.shape[1]):
            K = self.kernel.eval(self.kernel.params,
                                 np.expand_dims(np.unique(self.X[:, i]), 1))
            Ks.append(K)

        f_pred = kron_mvp(Ks, kron_mvp(self.K_invs, self.q_mu))

        return f_pred


class MFSVI(SVIBase):

    def __init__(self, kernel, likelihood, X, y,
                 mu=None, obs_idx=None, opt_kernel=False):
        """
        Args:
            kernel (): kernel function
            likelihood (): likelihood function. Requires log_like(),
             grad(), and hess()
            functions
            X (): data
            y (): responses
            mu (): prior mean
            noise (): noise variance
            obs_idx (): if dealing with partial grid, indices of grid
             that are observed
            verbose (): print or not
        """

        super(MFSVI, self).__init__(kernel, likelihood, X, y,
                                    mu, obs_idx, opt_kernel)

        self.q_S = np.ones(self.n) *\
            np.log(self.Ks[0][0, 0] ** self.d)
        self.v_mu, self.v_s, self.v_k, self.m_mu, \
            self.m_s, self.m_k = (None for _ in range(6))

    def run(self, its, n_samples=1, notebook_mode=True):
        """
        Runs stochastic variational inference
        Args:
            its (int): Number of iterations
            n_samples (int): Number of samples for SVI
        Returns: Nothing, but updates instance variables
        """

        if notebook_mode:
            t = tqdm_notebook(range(its), leave=False)
        else:
            t = trange(its, leave=False)

        for i in t:

            KL_grad_S, KL_grad_mu = self.grad_KL_S(),\
                                    self.grad_KL_mu()
            grads_mu, grads_S, es, rs = ([] for i in range(4))

            for j in range(n_samples):
                eps = np.random.normal(size=self.n)
                r = self.q_mu + np.multiply(np.sqrt(np.exp(self.q_S)),
                                            eps)
                like_grad_S, like_grad_mu = self.grad_like(r, eps)

                grad_mu = np.clip(-KL_grad_mu + like_grad_mu,
                                  -self.max_grad, self.max_grad)
                grad_S = np.clip(-KL_grad_S + like_grad_S,
                                 -self.max_grad, self.max_grad)

                grads_mu.append(grad_mu)
                grads_S.append(grad_S)
                es.append(eps)
                rs.append(r)

            if i % 50 == 0:
                obj, kl, like = self.eval_obj(self.q_S, self.q_mu, rs)
                self.elbos.append(-obj)

            t.set_description("ELBO: " + '{0:.2f}'.format(-obj) +
                              " | KL: " + '{0:.2f}'.format(kl) +
                              " | logL: " + '{0:.2f}'.format(like))

            S_vars = (self.q_S, np.mean(grads_S, 0))
            mu_vars = (self.q_mu, np.mean(grads_mu, 0))
            kern_and_grad = None

            if self.opt_kernel:
                kern_grad = self.grad_kern()
                if kern_grad is None:
                    print "kernel gradient failed"
                    return

                kern_grad_clip = np.clip(kern_grad, -self.max_grad,
                                         self.max_grad)
                kern_and_grad = (self.kernel.params, kern_grad_clip)

            self.q_mu, self.m_mu, self.v_mu = self.optimizer.step(mu_vars, self.m_mu,
                                                        self.v_mu, i+1)
            self.q_S, self.m_s, self.v_s = self.optimizer.step(S_vars, self.m_s,
                                                     self.v_s, i+1)
            if self.opt_kernel:
                self.kernel.params, self.m_k, self.v_k =\
                    self.optimizer.step(kern_and_grad,
                                        self.m_k, self.v_k, i+1)
                self.Ks, self.K_invs = self.construct_Ks()

            if i > its/2 and self.loss_check() == True:
                print("converged at", i, "iterations")
                return

        return

    def eval_obj(self, S, q_mu, rs, kern_params=None):
        """
        Evaluates variational objective
        Args:
            Rs (): Variational covariances
             (Cholesky decomposition of Kronecker decomp)
            q_mu (): Variational mean
            r (): Transformed random sample
        Returns: ELBO evaluation
        """
        kl = self.KLqp(S, q_mu, kern_params)
        like = 0.

        for r in rs:

            if self.obs_idx is not None:
                r_obs = r[self.obs_idx]
            else:
                r_obs = r

            like += np.sum(self.likelihood.log_like(r_obs, self.y))/len(rs)

        obj = kl - like

        return obj, kl, like

    def line_search(self, S_grads, mu_grads, obj_init, r, eps):
        """
        Performs line search to find optimal step size
        Args:
            Rs_grads (): Gradients of R (variational covariances)
            mu_grads (): Gradients of mu (variational mean)
            obj_init (): Initial objective value
            r (): transformed random Gaussian sample
            eps (): random Gaussian sample
        Returns: Optimal step size
        """
        step = 1.

        while step > 1e-9:

            S_search = [S + step * S_grad
                        for S_grad, S in S_grads]
            mu_search = mu_grads[1] + step * mu_grads[0]
            r_search = mu_search + np.multiply(S_search, eps)
            obj_search, kl_search, like_search = self.eval_obj(S_search, mu_search,
                                                               r_search)
            if obj_init - obj_search > step:
                return S_search, mu_search, obj_search, step

            step = step * 0.5

        return None

    def KLqp(self, S, q_mu, kern_params):
        """
        Calculates KL divergence between q and p
        Args:
            Rs (): Variational covariance
            q_mu (): Variational mean
        Returns: KL divergence between q and p
        """

        if kern_params is None:
            K_invs = self.K_invs
            k_inv_diag = self.k_inv_diag
            det_K = self.det_K
        else:
            Ks, K_invs = self.construct_Ks()
            k_inv_diag = kron_list_diag(K_invs)
            det_K = self.log_det_K(Ks)

        k_inv_mu = kron_mvp(K_invs, self.mu - q_mu)
        mu_penalty = np.sum(np.multiply(self.mu - q_mu, k_inv_mu))
        det_S = np.sum(S)
        trace_term = np.sum(np.multiply(k_inv_diag, np.exp(S)))

        kl = 0.5 * (det_K - self.n - det_S +
                    trace_term + mu_penalty)

        return kl

    def grad_KL_S(self):
        """
        Gradient of KL divergence w.r.t variational covariance
        Returns: returns gradient
        """
        euc_grad = 0.5 * (-1. + np.multiply(self.k_inv_diag, np.exp(self.q_S)))

        return euc_grad

    def grad_KL_mu(self):
        """
        Gradient of KL divergence w.r.t variational mean
        Returns: returns gradient
        """
        return -kron_mvp(self.K_invs, self.mu - self.q_mu)

    def grad_like(self, r, eps):
        """
        Gradient of likelihood w.r.t variational parameters
        Args:
            r (): Transformed random sample
            eps (): Random sample
        Returns: gradient w.r.t covariance, gradient w.r.t mean
        """
        if self.obs_idx is not None:
            r_obs = r[self.obs_idx]
        else:
            r_obs = r

        dr = self.likelihood_opt(r_obs, self.y)
        dr = np.nan_to_num(dr)

        if self.obs_idx is not None:
            grad_mu = np.zeros(self.n)
            grad_mu[self.obs_idx] = dr
        else:
            grad_mu = dr
        grad_S = np.multiply(grad_mu,
                             np.multiply(eps,
                                         np.multiply(0.5/np.sqrt(
                                                np.exp(self.q_S)),
                                                np.exp(self.q_S))))

        return grad_S, grad_mu

    def grad_kern(self):
        """

        Returns: Gradient of KL w.r.t base kernel parameters

        """

        k_inv_mu = self.q_mu - self.mu
        grads = []
        term1 = np.ones(len(self.kernel.params))
        term2 = np.expand_dims(np.ones(len(self.kernel.params)), 1)

        for i, X in enumerate(reversed(self.X_dims)):

            grad = self.kernel_opt(self.kernel.params, X[0], X)
            print grad.shape
            toep_grad = np.stack([toeplitz(grad[:, :, k])
                                  for k in range(grad.shape[2])],
                                 axis=-1)
            grads.append(toep_grad)

            term1 = np.multiply(term1,
                                np.einsum('ij, ijk -> k',
                                          self.K_invs[i], toep_grad))

            diag = np.einsum('ij,jik->jk', self.K_invs[i],
                             np.einsum('ijk,j...', toep_grad,
                                       self.K_invs[i]))
            term2 = np.stack([np.hstack([ii * term2[d, :]
                              for ii in diag[:, d]])
                              for d in range(diag.shape[1])])

            k_inv_mu = np.reshape(k_inv_mu, [X.shape[0], -1])
            k_inv_mu = np.dot(self.K_invs[i], k_inv_mu).T

        term2 = np.sum(np.multiply(term2.T,
                                   np.expand_dims(np.exp(self.q_S), 1)),
                       0)
        k_inv_mu = np.reshape(k_inv_mu, [-1])
        term3 = np.tile(k_inv_mu, (len(self.kernel.params), 1)).T

        for grad in grads:
            term3 = np.reshape(term3, [grad.shape[0], -1,
                                       grad.shape[2]])
            term3 = np.stack([np.dot(grad[:, :, d], term3[:, :, d]).T
                              for d in range(len(self.kernel.params))],
                             axis=-1)

        term3 = np.reshape(term3,
                           [-1, len(self.kernel.params)])
        term3 = np.sum(np.multiply(np.expand_dims(k_inv_mu, 1),
                                   term3), 0)

        return -0.5*(term1 - term2 - term3)

    def predict(self):
        """
        GP predictions
        Returns: predictions
        """
        Ks = [self.kernel.eval(self.kernel.params, X_dim)
              for X_dim in self.X_dims]
        K_invs = [np.linalg.inv(K + np.diag(np.ones(K.shape[0]))*self.kernel.params[2])
                  for K in Ks]

        f_pred = kron_mvp(Ks, kron_mvp(K_invs, self.q_mu))

        return f_pred

    def sample_post(self, n_samples=1):

        return self.q_mu + \
               np.multiply(np.expand_dims(
                              np.sqrt(np.exp(self.q_S)), 1),
                           np.random.normal(size=(self.n, n_samples))).flatten()

    def loss_check(self):

        if sum(x >= y for x, y in zip(self.elbos[-10:], self.elbos[-9:])) > 50 and \
                self.elbos[-1] - self.elbos[-10] < 1e-3 * abs(self.elbos[-10]):
            return True


class FullSVI(SVIBase):


    def __init__(self, kernel, likelihood, X, y, mu = None, obs_idx=None):
        """
        Args:
            kernel (GPy.Kernel): kernel function
            likelihood (): likelihood function. Requires log_like(), grad(), and hess()
            functions
            X (): data
            y (): responses
            mu (): prior mean
            noise (): noise variance
            obs_idx (): if dealing with partial grid, indices of grid that are observed
            verbose (): print or not
        """

        super(FullSVI, self).__init__(kernel, likelihood, X, y, mu, obs_idx)

        self.Rs = self.initialize_Rs_prior()
        self.trace_term, self.traces = self.calc_trace_term()

    def run(self, its):
        """
        Runs stochastic variational inference
        Args:
            its (): Number of iterations
        Returns: Nothing, but updates instance variables
        """

        t = trange(its, leave=True)

        for i in t:
            self.trace_term, self.traces = self.calc_trace_term()
            KL_grad_R = self.grad_KL_R()
            KL_grad_mu = self.grad_KL_mu()

            eps = np.random.normal(size = self.n)
            r = self.q_mu + kron_mvp(self.Rs, eps)
            like_grad_R, like_grad_mu = self.grad_like(r, eps)

            grad_R = [np.clip(-KL_grad_R[i] + like_grad_R[i], -1e3, 1e3)
                      for i in range(len(KL_grad_R))]
            grad_mu= np.clip(-KL_grad_mu + like_grad_mu, -1e3, 1e3)
            R_and_grads = zip(grad_R, self.Rs)
            mu_and_grad = (grad_mu, self.q_mu)

            obj, kl, like = self.eval_obj(self.Rs, self.q_mu, r)
            self.elbos.append(-obj)

            ls_res = self.line_search(R_and_grads, mu_and_grad, obj, r, eps)

            step = 0.
            if ls_res is not None:
                step = ls_res[-1]

            t.set_description("ELBO: " + '{0:.2f}'.format(-obj) +
                              " | KL: " + '{0:.2f}'.format(kl) +
                              " | logL: " + '{0:.2f}'.format(like) +
                              " | step: " + str(step))
            if ls_res is not None:
                self.Rs = ls_res[0]
                self.q_mu = ls_res[1]

        return

    def eval_obj(self, Rs, q_mu, r):
        """
        Evaluates variational objective
        Args:
            Rs (): Variational covariances
             (Cholesky decomposition of Kronecker decomp)
            q_mu (): Variational mean
            r (): Transformed random sample
        Returns: ELBO evaluation
        """
        kl = self.KL_calc(Rs, q_mu)
        if self.obs_idx is not None:
            r_obs = r[self.obs_idx]
        else:
            r_obs = r
        like = np.sum(self.likelihood.log_like(r_obs, self.y))
        obj = kl - like

        return obj, kl, like

    def KL_calc(self, Rs, q_mu):
        """
        Calculates KL divergence between q and p
        Args:
            Rs (): Variational covariance
            q_mu (): Variational mean
        Returns: KL divergence between q and p
        """
        k_inv_mu = kron_mvp(self.K_invs, self.mu - q_mu)
        mu_penalty = np.sum(np.multiply(self.mu -q_mu, k_inv_mu))
        det_S = self.log_det_S(Rs)
        trace_term = self.calc_trace_term(Rs)[0]
        kl = 0.5 * (self.det_K - self.n - det_S +
                      trace_term + mu_penalty)

        if kl < 0:
            return 0.

        return max(0, kl)

    def grad_KL_R(self):
        """
        Gradient of KL divergence w.r.t variational covariance
        Returns: returns gradient
        """
        return [np.diag(-2*self.n/self.Rs[d].shape[0]/
                         np.diag(self.Rs[d])) +\
                         np.prod(self.traces)/self.traces[d] *
                         np.dot(self.K_invs[d], self.Rs[d])
                         for d in range(len(self.Rs))]

    def grad_KL_mu(self):
        """
        Gradient of KL divergence w.r.t variational mean
        Returns: returns gradient
        """
        return -kron_mvp(self.K_invs, self.mu - self.q_mu)

    def grad_like(self, r, eps):
        """
        Gradient of likelihood w.r.t variational parameters
        Args:
            r (): Transformed random sample
            eps (): Random sample
        Returns: gradient w.r.t covariance, gradient w.r.t mean
        """
        if self.obs_idx is not None:
            r_obs = r[self.obs_idx]
        else:
            r_obs = r


        dr = self.likelihood_opt(r_obs, self.y)
        dr[np.isnan(dr)] = 0.
        grads_R = []

        for d in range(len(self.Rs)):

            Rs_copy = deepcopy(self.Rs)
            n = Rs_copy[d].shape[0]
            grad_R = np.zeros((n, n))

            for i, j in zip(*np.triu_indices(n)):
                R_d = np.zeros((n, n))
                R_d[i, j] = 1.
                Rs_copy[d] = R_d
                dR_eps = kron_mvp(Rs_copy, eps)
                if self.obs_idx is not None:
                    dR_eps = dR_eps[self.obs_idx]
                grad_R[i, j] = np.sum(np.multiply(dr, dR_eps))
            grads_R.append(grad_R)

        if self.obs_idx is not None:
            grad_mu = np.zeros(self.n)
            grad_mu[self.obs_idx] = dr
        else:
            grad_mu = dr

        return grads_R, grad_mu

    def line_search(self, Rs_grads, mu_grads, obj_init, r, eps):
        """
        Performs line search to find optimal step size
        Args:
            Rs_grads (): Gradients of R (variational covariances)
            mu_grads (): Gradients of mu (variational mean)
            obj_init (): Initial objective value
            r (): transformed random Gaussian sample
            eps (): random Gaussian sample
        Returns: Optimal step size
        """
        step = 1.

        while step > 1e-9:

            R_search = [R + step * R_grad
                        for R_grad, R in Rs_grads]
            mu_search = mu_grads[1] + step * mu_grads[0]
            r_search = mu_search + kron_mvp(R_search, eps)
            obj_search, kl_search, like_search = self.eval_obj(R_search, mu_search,
                                                               r_search)
            if obj_init - obj_search > step:
                return R_search, mu_search, obj_search, step

            step = step * 0.5

        return None

    def calc_trace_term(self, Rs = None):
        """
        Calculates trace term for objective function
        Args:
            Rs (): trace of variational covariance, and individual trace over dimensions
        Returns:
        """
        if Rs is None:
            Rs = self.Rs

        traces = [np.trace(np.dot(self.K_invs[d],
                                     np.dot(np.transpose(Rs[d]), Rs[d])))
                                     for d in range(len(self.K_invs))]

        return np.prod(traces), traces

    def log_det_S(self, Rs = None):
        """
        Log determinant of variational covariance
        Args:
            Rs (): Kronecker decomposed variational covariance
        Returns: determinant
        """
        if Rs is None:
            Rs = self.Rs

        return 2*np.sum([self.n/R.shape[0]*
                                 np.sum(np.log(np.diag(R)))
                                 for R in Rs])

    def initialize_Rs(self):
        """
        Initializes upper triangular decomp
         of kronecker decomp of vairational covariance
        using identity matrix
        Returns: Rs (identity matrices)
        """
        return[np.eye(K.shape[0])
            for K in self.Ks]

    def initialize_Rs_prior(self):
        """
        Initializes Rs using cholesky decomps of prior covariance
        Returns: Rs (cholesky decomp of prior covariances)
        """
        return [np.transpose(np.linalg.cholesky(K))
                for K in self.Ks]


    def cg_prod(self, A, x):
        """
        Not currently used, but for conjugate gradient method
        Args:
            A ():
            x ():
        Returns:
        """
        return kron_mvp(A, x)

    def sample_post(self):
        """
        Draws a sample from the GPLVM posterior
        Returns: sample
        """

        eps = np.random.normal(size = self.n)
        r = self.q_mu + kron_mvp(self.Rs, eps)

        return r


import autograd.numpy as np

"""
basic likelihood class
"""

class Poisson:
    """
    Implements Poisson likelihood with exponential link
    """
    def log_like(self, log_rate, y):

        return np.multiply(y, log_rate) - np.exp(log_rate)

class Bernoulli:

    """
    Implements Bernoulli likelihood with logistic sigmoid link
    """
    def log_like(self, g, y):

        return -np.multiply(y, np.log(1 + np.exp(-g))) - \
                              np.multiply(1 - y, np.log(1 + np.exp(g)))




import GPy
import numpy as np
import itertools
from scipy.stats import rankdata

"""
basic utilities for simulating data
"""


def sim_f(X, kernel, mu=3):
    """
    simulates function values given X
    Args:
        X (np.array): data points
        k (GPy.kernel): kernel function
        mu (np.array): prior mean

    Returns: sampled function values

    """
    return np.random.multivariate_normal(np.ones(X.shape[0]) * mu,
                                         kernel.eval(kernel.params, X))

def sim_X(D=2, N_dim=30, lower=0, upper=100):
    """
    Simulates X on a rectilinear grid
    Args:
        D (int): dimensions
        N_dim (int): number of points per dimension
        lower (float): lower bound for uniform draw
        upper (float): upper bound for uniform draw

    Returns: points on a grid

    """
    grid = [np.sort(np.random.uniform(lower, upper, size=N_dim))
            for d in range(D)]

    return np.array(list(itertools.product(*grid)))


def sim_X_equispaced(D=2, N_dim=20, lower=0, upper=100):

    grid = [np.arange(lower, upper, (upper-lower)*1.0/N_dim) for d in range(D)]

    return np.array(list(itertools.product(*grid)))

def poisson_draw(f, noise_val):
    """

    Args:
        f (np.array): draws a poisson based on function values
                      (with some noise added)
        noise_val(float): between zero and one, add normal noise to f

    Returns: poisson draws

    """
    return np.random.poisson(np.exp(f + noise_val * np.random.normal(0, 1,
                             size=len(f))))


def rand_partial_grid(X, y, prop):

    indices = np.sort(np.random.choice(X.shape[0], int(X.shape[0] * prop),
                      replace=False))
    X_partial = X[indices]
    y_partial = y[indices]
    X_partial = X_partial[np.lexsort((X_partial[:, 1], X_partial[:, 0]))]

    return X_partial, y_partial

def fill_grid(X, y):
    """
    Fills a partial grid with "imaginary" observations
    Args:
        X (np.array): data that lies on a partial grid

    Returns:
        X_grid: full grid X (including real and imagined points)
        y_full: full grid y (with zeros corresponding to imagined points)
        obs_idx: indices of observed points
        imag_idx: indices of imagined points

    """

    D = X.shape[1]
    x_dims = [np.unique(X[:, d]) for d in range(D)]

    X_grid = np.array(list(itertools.product(*x_dims)))

    d_indices = [{k: v for k, v in zip(x_dims[d], range(x_dims[d].shape[0]))}
                 for d in range(D)]
    grid_part = np.ones([x_d.shape[0] for x_d in x_dims])*-1

    for i in range(X.shape[0]):
        idx = tuple([d_indices[d][X[i, d]] for d in range(D)])
        grid_part[idx] = 1

    obs_idx = np.where(grid_part.flatten() > -1)[0]
    imag_idx = np.where(grid_part.flatten() == -1)[0]

    y_full = np.zeros(X_grid.shape[0])
    y_full[obs_idx] = y

    return X_grid, y_full, obs_idx, imag_idx

def min_grid(X, y):

    grid_size = len(np.unique(X[:,0]))
    indices = []

    grid_idx = [np.random.permutation(range(np.ceil(np.power(X.shape[0],
                                      1./X.shape[1])).astype(np.int32)))
                for _ in range(X.shape[1])]
    grid_idx = zip(*grid_idx)

    for g in grid_idx:
        i = 0
        for d in xrange(X.shape[1]):
            i+= g[d]*grid_size**(X.shape[1] - d - 1)
        indices.append(i)

    X_partial = X[indices]
    y_partial = y[indices]
    X_partial = X_partial[np.lexsort(tuple(X_partial[:, d]
                                      for d in reversed(range(X.shape[1]))))]

    return X_partial, y_partial


import sys
from tqdm import trange
import autograd.numpy as np
from autograd import elementwise_grad as egrad
from gp3.utils.optimizers import CG
from gp3.utils.structure import kron_list, kron_mvp
from scipy.linalg import toeplitz
from scipy.optimize import minimize


"""
Class for Kronecker inference of GPs. Inspiration from GPML.

For references, see:

Flaxman and Wilson (2014),
Fast Kronecker Inference in Gaussian Processes with non-Gaussian Likelihoods

Rassmussen and Williams (2006),
Gaussian Processes for Machine Learning

Wilson et al (2012),
Fast Kernel Learning for Multidimensional Pattern Extrapolation

Wilson et al (2014),
Thoughts on Massively Scalable Gaussian Processes

Most of the notation follows R and W chapter 2, and Flaxman and Wilson

"""


class Laplace:

    def __init__(self, kernel, likelihood, X, y, mu = None,
                 tau=0.5, obs_idx=None, verbose=False, noise = 1e-6):
        """

        Args:
            kernel (kernels.Kernel): kernel function to use for inference
            likelihood (likelihoods.Likelihood): likelihood
            X (np.array): data
            y (np.array): output
            tau (float): Newton line search hyperparam
            obs_idx (np.array): Indices of observed points (partial grid)
            verbose (bool): verbose or not
        """

        self.verbose = verbose
        self.X = X
        self.y = y
        self.n = self.X.shape[0]
        self.d = self.X.shape[1]
        self.X_dims = [np.expand_dims(np.unique(X[:,i]), 1) for i in range(self.d)]
        if mu is None:
            self.mu = np.zeros(self.n)
        else:
            self.mu = mu
        self.obs_idx = obs_idx

        self.kernel = kernel
        self.noise = noise
        self.likelihood = likelihood
        self.Ks = self.construct_Ks()
        self.K_eigs = [np.linalg.eig(K) for K in self.Ks]
        self.root_eigdecomp = None

        self.alpha = np.zeros([X.shape[0]])
        self.W = np.zeros([X.shape[0]])
        self.grads = np.zeros([X.shape[0]])
        self.opt = CG(self.cg_prod)

        self.f = self.mu
        self.f_pred = self.f
        self.tau = tau

        self.grad_func = egrad(self.likelihood.log_like)
        self.hess_func = egrad(self.grad_func)


    def construct_Ks(self, kernel=None):
        """
        Constructs kronecker-decomposed kernel matrix
        Args:
            kernel (): kernel (if not using kernel passed in constructor)
        Returns: Rist of kernel evaluated at each dimension
        """

        if kernel is None:
            kernel = self.kernel

        Ks = [toeplitz(kernel.eval(kernel.params, X_dim[0], X_dim)) +\
            np.diag(np.ones(X_dim.shape[0])*self.noise) for X_dim in self.X_dims]

        return Ks

    def sqrt_eig(self):
        """
        Calculates square root of kernel matrix using
         fast kronecker eigendecomp.
        This is used in stochastic approximations
         of the predictive variance.

        Returns: Square root of kernel matrix

        """
        res = []

        for e, v in self.K_eigs:
            e_root_diag = np.sqrt(e)
            e_root = np.diag(np.real(np.nan_to_num(e_root_diag)))
            res.append(np.real(np.dot(np.dot(v, e_root), np.transpose(v))))

        res = np.squeeze(kron_list(res))
        self.root_eigdecomp = res

        return res

    def run(self, max_it):
        """
        Runs Kronecker inference. Updates instance variables.

        Args:
            max_it (int): maximum number of iterations.

        Returns: max iterations, iteration number, objective

        """
        if self.obs_idx is not None:
            k_diag = np.ones(self.X.shape[0]) * 1e12
            k_diag[self.obs_idx] = self.noise
            self.k_diag = k_diag
            self.precondition = np.clip(1.0 / np.sqrt(self.k_diag),
                                                 0, 1)
        else:
            self.k_diag = None
            self.precondition = None

        delta = sys.float_info.max
        it = 0

        t = trange(max_it)

        for i in t:
            max_it, it, delta, step, psi = self.step(max_it, it, delta)
            t.set_description("Objective: " + '{0:.2f}'.format(psi) +
                              " | Step Size: " + '{0:.2f}'.format(step))
            if delta < 1e-9:
                break

        self.f_pred = kron_mvp(self.Ks, self.alpha) + self.mu
        self.update_derivs()

        return

    def step(self, max_it, it, delta):
        """
        Runs one step of Kronecker inference
        Args:
            max_it (int): maximum number of Kronecker iterations
            it (int): current iteration
            delta (np.array): change in step size

        Returns: max iteration, current iteration, previous objective,
         change in objective

        """

        self.f = kron_mvp(self.Ks, self.alpha) + self.mu
        if self.k_diag is not None:
            self.f += np.multiply(self.alpha, self.k_diag)
        psi = self.log_joint(self.f, self.alpha)
        self.update_derivs()

        b = np.multiply(self.W, self.f - self.mu) + self.grads
        if self.precondition is not None:
            z = self.opt.cg(self.Ks, np.multiply(self.precondition,
                            np.multiply(1.0/np.sqrt(self.W), b)))
        else:
            z = self.opt.cg(self.Ks, np.multiply(1.0/np.sqrt(self.W), b))

        delta_alpha = np.multiply(np.sqrt(self.W), z) - self.alpha
        step_size = self.line_search(delta_alpha, psi, 20)

        delta = step_size

        if delta > 1e-9:
            self.alpha = self.alpha + delta_alpha*step_size
            self.alpha = np.where(np.isnan(self.alpha),
                                  np.ones_like(self.alpha) * 1e-9, self.alpha)

        it = it + 1

        return max_it, it, delta, step_size, psi

    def line_search(self, delta_alpha, obj_prev, max_it, verbose = False):
        """
        Executes line search for optimal Newton step
        Args:
            delta_alpha (np.array): change in search direction
            obj_prev (np.array): previous objective value
            max_it (int): maximum number of iterations

        Returns: optimal step size

        """
        obj_search = sys.float_info.max
        min_obj = obj_prev
        step_size = 2.0
        opt_step = 0.0
        t = 1

        while t < max_it and obj_prev - obj_search < step_size * t:
            obj_prev, min_obj, delta_alpha, step_size, max_it, t, opt_step = \
                self.search_step(obj_prev, min_obj, delta_alpha, step_size,
                             max_it, t, opt_step)
        return opt_step


    def search_step(self, obj_prev, min_obj, delta_alpha,
                    step_size, max_it, t, opt_step):
        """
        Executes one step of a backtracking line search
        Args:
            obj_prev (np.array): previous objective
            obj_search (np.array): current objective
            min_obj (np.array): current minimum objective
            delta_alpha (np.array): change in step size
            step_size (np.array): current step size
            max_it (int): maximum number of line search iterations
            t (np.array): current line search iteration
            opt_step (np.array): optimal step size until now

        Returns: updated parameters
        """
        alpha_search = np.squeeze(self.alpha + step_size * delta_alpha)
        f_search = np.squeeze(kron_mvp(self.Ks, alpha_search)) + self.mu

        if self.k_diag is not None:
            f_search += np.multiply(self.k_diag, alpha_search)

        obj_search = self.log_joint(f_search, alpha_search)

        if min_obj > obj_search:
            opt_step = step_size
            min_obj = obj_search
        step_size = self.tau * step_size

        t = t + 1

        return obj_prev, min_obj, delta_alpha,\
            step_size, max_it, t, opt_step

    def log_joint(self, f=None, alpha=None):

        """
        Evaluates objective function (negative log likelihood plus GP penalty)
        Args:
            f (): function values (if not same as class variable)
            alpha (): alpha (if not same as class variable)

        Returns:
        """

        if self.obs_idx is not None:
            f_lim = f[self.obs_idx]
            alpha_lim = alpha[self.obs_idx]
            mu_lim = self.mu[self.obs_idx]
            like = -np.sum(self.likelihood.log_like(f_lim, self.y))
            like = np.where(np.isnan(like), np.ones_like(like)*1e12, like)
            return like +\
                0.5 * np.sum(np.multiply(alpha_lim, f_lim - mu_lim))

        return -np.sum(self.likelihood.log_like(f, self.y)) +\
            0.5 * np.sum(np.multiply(alpha, f - self.mu))

    def marginal(self, kernel):
        """
        calculates marginal likelihood
        Args:
            Ks_new: new covariance if needed
        Returns: np.array for marginal likelihood

        """

        if kernel.params is not None:
            self.Ks = self.construct_Ks(self.kernel)
            self.alpha = np.zeros([self.X.shape[0]])
            self.W = np.zeros([self.X.shape[0]])
            self.grads = np.zeros([self.X.shape[0]])
            self.f = self.mu
            self.f_pred = self.f
            self.run(10)

        Ks = self.Ks
        eigs = [np.expand_dims(np.linalg.eig(K)[0], 1) for K in Ks]
        eig_K = np.squeeze(kron_list(eigs))
        self.eig_K = eig_K

        if self.obs_idx is not None:
            f_lim = self.f[self.obs_idx]
            alpha_lim = self.alpha[self.obs_idx]
            mu_lim = self.mu[self.obs_idx]
            W_lim = self.W[self.obs_idx]
            eig_k_lim = eig_K[self.obs_idx]

            pen = -0.5 * np.sum(np.multiply(alpha_lim,
                                       f_lim - mu_lim))
            pen = np.where(np.isnan(pen), np.zeros_like(pen), pen)
            eigs = 0.5 * np.sum(np.log(1 + np.multiply(eig_k_lim,
                                       W_lim)))
            eigs = np.where(np.isnan(eigs), np.zeros_like(eigs), eigs)
            like = np.sum(self.likelihood.log_like(f_lim, self.y))
            like = np.where(np.isnan(like), np.zeros_like(like), like)

            return pen+eigs+like

        pen = -0.5 * np.sum(np.multiply(self.alpha,
                                   self.f - self.mu))
        eigs = - 0.5*np.sum(np.log(1 +
                                   np.multiply(eig_K, self.W)))
        like = np.sum(self.likelihood.log_like(self.f, self.y))

        return -(pen+eigs+like)

    def variance(self, n_s):
        """
        Stochastic approximator of predictive variance.
         Follows "Massively Scalable GPs"
        Args:
            n_s (int): Number of iterations to run stochastic approximation

        Returns: Approximate predictive variance at grid points

        """

        if self.root_eigdecomp is None:
            self.root_eigdecomp = self.sqrt_eig()

        WK = np.dot(np.diag(np.sqrt(self.W)), self.root_eigdecomp)
        W_kd = None

        if self.precondition is not None:
            W_kd = np.multiply(np.sqrt(self.W), np.sqrt(self.k_diag))

        var = np.zeros([self.n])

        for i in range(n_s):
            g_m = np.random.normal(size = self.n)
            g_n = np.random.normal(size = self.n)
            if self.precondition is None:
                right_side = np.dot(WK, g_m) + g_n
            else:
                cov_term = np.dot(WK, g_m)
                noise_term = np.multiply(W_kd, g_n)
                right_side = np.multiply(self.precondition,
                                         cov_term + noise_term)
            right_side = np.nan_to_num(right_side)
            r = self.opt.cg(self.Ks, right_side)
            var += np.square(np.squeeze(kron_mvp(self.Ks,
                                        np.multiply(np.sqrt(self.W), r))))

        return np.clip(np.squeeze(self.kernel.eval(self.kernel.params,
            np.array([[0.]]), np.array([[0.]]))) -
                          var/n_s*1.0, 0., 1e3)

    def predict_mean(self, x_new):

        k_dims = [self.kernel.eval(self.kernel.params,
                                   np.expand_dims(np.unique(self.X[:, d]), 1),
                                   np.expand_dims(x_new[:, d], 1))
                  for d in self.X.shape[1]]
        kx = np.squeeze(kron_list(k_dims))
        mean = np.sum(np.multiply(kx, self.alpha)) + self.mu[0]

        return mean

    def cg_prod(self, Ks, p):
        """

        Args:
            p (tfe.Variable): potential solution to linear system

        Returns: product Ap (left side of linear system)

        """

        if self.precondition is None:
            return p + np.multiply(np.sqrt(self.W),
                                   kron_mvp(Ks,
                                            np.multiply(np.sqrt(self.W), p)))

        Cp = np.multiply(self.precondition, p)
        noise = np.multiply(np.multiply(self.precondition,
                                        np.multiply(self.W, self.k_diag)), Cp)
        wkw = np.multiply(np.multiply(self.precondition, np.sqrt(self.W)),
                          kron_mvp(Ks, np.multiply(np.sqrt(self.W), Cp)))

        return noise + wkw + np.multiply(self.precondition, Cp)

    def gather_derivs(self):
        """

        Returns: sum of gradients, hessians

        """

        obs_f = self.f[self.obs_idx]
        obs_grad = self.grad_func(obs_f, self.y)
        obs_hess = self.hess_func(obs_f, self.y)
        self.obs_hess = obs_hess

        agg_grad = np.zeros(self.n, np.float64)
        agg_hess = np.zeros(self.n, np.float64)

        for i, j in enumerate(self.obs_idx):
            agg_grad[j] += obs_grad[i]
            agg_hess[j] += obs_hess[i]

        return agg_grad, agg_hess

    def update_derivs(self):

        if self.obs_idx is None:
            self.grads = self.grad_func(self.f, self.y)
            hess = self.hess_func(self.f, self.y)
            self.hess = hess
            self.W = np.clip(-hess, 1e-9, 1e16)
        else:
            self.grads, hess = self.gather_derivs()
            self.hess = hess
            self.W = np.clip(-hess, 1e-9, 1e16)
        self.W = np.where(np.isnan(self.W), np.ones_like(self.W) * 1e-9,
                          self.W)

        return self.grads, self.hess

    def opt_kern(self):

        init_params = self.kernel.params

        return minimize(self.marginal, init_params, jac = False, method='CG')



from setuptools import setup

setup(
    name='gp3',
    packages = ['gp3', 'gp3.inference', 'gp3.likelihoods',
                'gp3.utils', 'gp3.kernels'],
    version='0.1.1',
    description='Gaussian Processes with Probabilistic Programming',
    author='Anuj Sharma',
    author_email="anuj.sharma@columbia.edu",
    install_requires=['numpy>=1.7',
                      'six>=1.10.0'],
    url='https://github.com/as4529/gp3',  # use the URL to the github repo
    extras_require={
        'notebooks': ['jupyter>=1.0.0'],
        'visualization': ['matplotlib>=1.3',
                          'plotly>=2.2.2',
                          'tqdm>=4.19.4']})